import sys

import psutil
import distributed
import inspect
import importlib
import importlib.util
import pathlib

import toolviper.utils.logger as logger
import toolviper.utils.console as console

from importlib.metadata import PackageNotFoundError, version

from packaging.version import parse as parse_version
from distributed.diagnostics.plugin import WorkerPlugin

from dask.widgets import get_template

from typing import Callable, Tuple, Dict, Any, Union

colorize = console.Colorize()

class MenrvaClient(distributed.Client):
    """
    This and extended version of the general Dask distributed client that will allow for
    plugin management and more extended features.
    """

    def _repr_html_(self):
        try:
            distributed.Client.current()

        except ValueError:
            logger.debug("<No Dask Client>")
            return None


        try:
            dle_version = parse_version(version("dask-labextension"))
            JUPYTERLAB = False if dle_version < parse_version("6.0.0") else True

        except PackageNotFoundError:
            JUPYTERLAB = False

        scheduler, info = self._get_scheduler_info()

        return get_template("client.html.j2").render(
            id=self.id,
            scheduler=scheduler,
            info=info,
            cluster=self.cluster,
            scheduler_file=self.scheduler_file,
            dashboard_link=self.dashboard_link,
            jupyterlab=JUPYTERLAB,
        )

    @staticmethod
    def thread_info()->Dict[str, Any]:

        try:
            client = distributed.Client.current()

        except ValueError:  # Using default Dask schedular.
            logger.warning("Couldn't find a current client instance, calculating thread information based on current system.")

            cpu_cores = psutil.cpu_count()
            total_memory = psutil.virtual_memory().total / (1024 ** 3)

            thread_info = {
                'n_threads': cpu_cores,
                'memory_per_thread': total_memory / cpu_cores
            }

            return thread_info

        memory_per_thread = -1
        n_threads = 0

        # client.cluster only exists for LocalCluster
        if client.cluster is None:
            worker_items = client.scheduler_info()['workers'].items()

        else:
            worker_items = client.cluster.scheduler_info['workers'].items()

        for worker_name, worker in worker_items:
            temp_memory_per_thread = (worker['memory_limit'] / worker['nthreads']) / (1024 ** 3)
            n_threads = n_threads + worker['nthreads']

            if (memory_per_thread == -1) or (memory_per_thread > temp_memory_per_thread):
                memory_per_thread = temp_memory_per_thread

        thread_info = {
            'n_threads': n_threads,
            'memory_per_thread': memory_per_thread
        }

        return thread_info

    @staticmethod
    def call(func: Callable, *args: Tuple[Any], **kwargs: Dict[str, Any]):
        try:
            params = inspect.signature(func).bind(*args, **kwargs)
            return func(*params.args, **params.kwargs)

        except TypeError as e:
            logger.error("There was an error calling the function: {}".format(e))

    @staticmethod
    def instantiate_module(
        plugin: str, plugin_file: str, *args: Tuple[Any], **kwargs: Dict[str, Any]
    ) -> WorkerPlugin:
        """

        Args:
            plugin (str): Name of plugin module.
            plugin_file (str): Name of module file. ** This should be moved into the module itself not passed **
            *args (tuple(Any)): This is any *arg that needs to be passed to the plugin module.
            **kwargs (dict[str, Any]): This is any **kwarg default values that need to be passed to the plugin module.

        Returns:
            Instance of plugin class.
        """
        spec = importlib.util.spec_from_file_location(plugin, plugin_file)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        for member in inspect.getmembers(module, predicate=inspect.isclass):
            plugin_instance = getattr(module, member[0])
            logger.debug("Loading plugin module: {}".format(plugin_instance))
            return MenrvaClient.call(plugin_instance, *args, **kwargs)

    def load_plugin(
        self,
        directory: str,
        plugin: str,
        name: str,
        *args: Union[Tuple[Any], Any],
        **kwargs: Union[Dict[str, Any], Any],
    ):
        plugin_file = ".".join((plugin, "py"))
        if pathlib.Path(directory).joinpath(plugin_file).exists():
            plugin_instance = MenrvaClient.instantiate_module(
                plugin=plugin,
                plugin_file="/".join((directory, plugin_file)),
                *args,
                **kwargs,
            )
            logger.debug(f"{plugin}")
            if sys.version_info.major == 3:
                if sys.version_info.minor > 8:
                    self.register_plugin(plugin_instance, name=name)

                else:
                    self.register_worker_plugin(plugin_instance, name=name)
            else:
                logger.warning("Python version may not be supported.")
        else:
            logger.error(
                "Cannot find plugins directory: {}".format(colorize.red(directory))
            )


def port_is_free(port):
    import socket
    import errno

    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

    try:
        s.bind(("127.0.0.1", port))
    except socket.error as e:
        if e.errno == errno.EADDRINUSE:
            logger.warning("Port is already in use.")
            return False
        else:
            # something else raised the socket.error exception
            logger.exception(e)
            return False

    logger.debug("Socket is free.")
    s.close()

    return True


def close_port(port):
    import psutil
    from psutil import process_iter
    from signal import SIGKILL

    for proc in process_iter():
        try:
            for conns in proc.connections(kind="inet"):
                if conns.laddr.port == port:
                    proc.send_signal(SIGKILL)
                    continue

        except psutil.AccessDenied:
            pass



import os

from importlib.metadata import version
from toolviper.utils.logger import setup_logger

__version__ = version("toolviper")

# Setup default logger instance for module
if not os.getenv("VIPER_LOGGER_NAME"):
    os.environ["VIPER_LOGGER_NAME"] = "toolviper"
    setup_logger(
        logger_name="toolviper",
        log_to_term=True,
        log_to_file=False,
        log_file="toolviper-logfile",
        log_level="INFO",
    )


import inspect

from dataclasses import dataclass
from typing import List, Union, Callable

PREVIOUS_FUNCTION = 1
PENULTIMATE_FUNCTION = 2


@dataclass
class ColorCodes:
    # Color formatting
    black: str = "\033[38;2;0;0;0m"
    grey: str = "\033[38;2;112;128;144m"

    red: str = "\033[38;2;220;20;60m"
    green: str = "\033[38;2;46;139;87m"
    yellow: str = "\033[38;2;245;200;30m"
    orange: str = "\033[38;2;255;160;0m"
    blue: str = "\033[38;2;50;50;205m"
    purple: str = "\033[38;2;128;05;128m"
    white: str = "\033[38;2;245;255;250m"

    # Text formatting
    bold: str = "\033[1m"
    faint: str = "\033[2m"
    italics: str = "\033[3m"
    underline: str = "\033[4m"
    blink: str = "\033[5m"
    highlight: str = "\033[7m"

    reset: str = "\033[0m"
    alert: str = "\033[7;38;2;220;60;20m"


class Colorize:
    def __init__(self):
        self.codes = ColorCodes()

        self.levels = {"info": self.blue}

    def bold(self, text: str) -> str:
        return "{format}{text}{reset}".format(
            format=self.codes.bold, text=text, reset=self.codes.reset
        )

    def faint(self, text: str) -> str:
        return "{format}{text}{reset}".format(
            format=self.codes.faint, text=text, reset=self.codes.reset
        )

    def italics(self, text: str) -> str:
        return "{format}{text}{reset}".format(
            format=self.codes.italics, text=text, reset=self.codes.reset
        )

    def underline(self, text: str) -> str:
        return "{format}{text}{reset}".format(
            format=self.codes.underline, text=text, reset=self.codes.reset
        )

    def blink(self, text: str) -> str:
        return "{format}{text}{reset}".format(
            format=self.codes.blink, text=text, reset=self.codes.reset
        )

    def highlight(self, text: str) -> str:
        return "{format}{text}{reset}".format(
            format=self.codes.highlight, text=text, reset=self.codes.reset
        )

    def white(self, text: str) -> str:
        return "{color}{text}{reset}".format(
            color=self.codes.white, text=text, reset=self.codes.reset
        )

    def black(self, text: str) -> str:
        return "{color}{text}{reset}".format(
            color=self.codes.black, text=text, reset=self.codes.reset
        )

    def grey(self, text: str) -> str:
        return "{color}{text}{reset}".format(
            color=self.codes.grey, text=text, reset=self.codes.reset
        )

    def red(self, text: str) -> str:
        return "{color}{text}{reset}".format(
            color=self.codes.red, text=text, reset=self.codes.reset
        )

    def green(self, text: str) -> str:
        return "{color}{text}{reset}".format(
            color=self.codes.green, text=text, reset=self.codes.reset
        )

    def yellow(self, text: str) -> str:
        return "{color}{text}{reset}".format(
            color=self.codes.yellow, text=text, reset=self.codes.reset
        )

    def orange(self, text: str) -> str:
        return "{color}{text}{reset}".format(
            color=self.codes.orange, text=text, reset=self.codes.reset
        )

    def blue(self, text: str) -> str:
        return "{color}{text}{reset}".format(
            color=self.codes.blue, text=text, reset=self.codes.reset
        )

    def purple(self, text: str) -> str:
        return "{color}{text}{reset}".format(
            color=self.codes.purple, text=text, reset=self.codes.reset
        )

    def alert(self, text: str) -> str:
        return "{color}{text}{reset}".format(
            color=self.codes.alert, text=text, reset=self.codes.reset
        )

    @staticmethod
    def from_ansi(
        color: Union[str, List],
        bold: bool = False,
        italics: bool = False,
        faint: bool = False,
        underline: bool = False,
        highlight: bool = False,
        blink: bool = False,
    ):
        args = locals()
        args.pop("color")

        escape_code = "\033["

        if True not in args.values():
            escape_code = "".join((escape_code, "0"))

        else:
            option_list = []

            if bold:
                option_list.append("1")

            if faint:
                option_list.append("2")

            if italics:
                option_list.append("3")

            if underline:
                option_list.append("4")

            if blink:
                option_list.append("5")

            if highlight:
                option_list.append("7")

            option_string = ";".join(option_list)
            escape_code = "".join((escape_code, option_string))

        escape_code = ";".join((escape_code, "38;2"))
        color_code = ";".join(map(str, color)) + "m"

        return ";".join((escape_code, color_code))

    def format(
        self,
        text: str,
        color: Union[List, str],
        bold: bool = False,
        italics: bool = False,
        faint: bool = False,
        underline: bool = False,
        highlight: bool = False,
        blink: bool = False,
    ):
        if isinstance(color, list):
            return "{color}{text}{reset}".format(
                color=self.from_ansi(
                    color=color,
                    bold=bold,
                    italics=italics,
                    faint=faint,
                    underline=underline,
                    highlight=highlight,
                    blink=blink,
                ),
                text=text,
                reset=self.codes.reset,
            )
        else:
            ascii_list = self.get_color_ascii_list(color=color)
            return "{color}{text}{reset}".format(
                color=self.from_ansi(
                    color=ascii_list,
                    bold=bold,
                    italics=italics,
                    faint=faint,
                    underline=underline,
                    highlight=highlight,
                    blink=blink,
                ),
                text=text,
                reset=self.codes.reset,
            )

    def get_color_function(self, color: str) -> Callable:
        import inspect

        for method in inspect.getmembers(self):
            if color is method[0]:
                return getattr(self, method[0])

        return self.black

    def get_color_ascii_list(self, color: str):
        import re
        import inspect

        for member in inspect.getmembers(self.codes):
            if member[0] == color:
                return re.split("[;m]", member[1])[-4:-1]


def add_verbose_info(message: str, color: str = "blue") -> str:
    function_name = inspect.stack()[PENULTIMATE_FUNCTION].function
    colorize = Colorize()
    color_function = colorize.get_color_function(color=color)

    return "[{function_name}]: {message}".format(
        function_name=color_function(function_name), message=message
    )



import sys
import click

import toolviper.utils.logger as logger
import distributed

from distributed.diagnostics.plugin import WorkerPlugin


class DaskWorker(WorkerPlugin):
    def __init__(self, *args, **kwargs):
        self.worker = None
        self.logger = None

        self.local_cache = True

        self.logger_name = "worker"
        self.log_to_term = True
        self.log_to_file = False
        self.log_file = None
        self.log_level = "INFO"

        if "local_cache" in kwargs:
            self.local_cache = kwargs["local_cache"]

        # Update parameters if needed.
        self.set_log_parameters(kwargs)

    def set_log_parameters(self, kwargs):
        for param in kwargs["log_params"]:
            if param in self.__dict__:
                setattr(self, param, kwargs["log_params"][param])

    def get_logger(self):
        return self.logger

    def setup(self, worker: distributed.worker.Worker):
        """
        Run when the plugin is attached to a worker. This happens when the plugin is registered
        and attached to existing workers, or when a worker is created after the plugin has been
        registered.
        """

        self.logger = logger.setup_worker_logger(
            logger_name=self.logger_name,
            log_to_term=self.log_to_term,
            log_to_file=self.log_to_file,
            log_file=self.log_file,
            log_level=self.log_level,
            worker=worker,
        )

        self.logger.debug(
            "Logger created on worker " + str(worker.id) + ",*," + str(worker.address)
        )

        # Documentation https://distributed.dask.org/en/stable/worker.html#distributed.worker.Worker
        self.worker = worker

        if self.local_cache:
            ip = worker.address[
                worker.address.rfind("/") + 1 : worker.address.rfind(":")
            ]

            self.logger.debug(str(worker.id) + ",*," + ip)

            worker.state.available_resources = {
                **worker.state.available_resources,
                **{ip: 1},
            }


# https://github.com/dask/distributed/issues/4169
@click.command()
@click.option("--local_cache", default=False)
@click.option("--log_to_term", default=True)
@click.option("--log_to_file", default=False)
@click.option("--log_file", default="worker-")
@click.option("--log_level", default="INFO")
async def dask_setup(
    worker: distributed.worker.Worker,
    local_cache: str,
    log_to_term: str,
    log_to_file: str,
    log_file: str,
    log_level: str,
):
    log_params = {
        "log_to_term": log_to_term,
        "log_to_file": log_to_file,
        "log_file": log_file,
        "log_level": log_level,
    }

    plugin = DaskWorker(local_cache, log_params)

    if sys.version_info.major == 3:
        if sys.version_info.minor > 8:
            await worker.client.register_plugin(plugin, name="worker_logger")

        else:
            await worker.client.register_worker_plugin(plugin, name="worker_logger")

    else:
        logger.warning("Python version may not be supported.")



from .console import Colorize, add_verbose_info
from .parameter import validate, set_config_directory
from .protego import Protego
from .logger import info, debug, warning, error, critical, get_logger, setup_logger

from .data import download

__submodules__ = ["data"]

__all__ = __submodules__ + [s for s in dir() if not s.startswith("_")]



import os
import re
import pathlib
import distributed

from toolviper.dask.client import local_client


class TestToolViperClient:
    @classmethod
    def setup_class(cls):
        """setup any state specific to the execution of the given test class
        such as fetching test data"""
        pass

    @classmethod
    def teardown_class(cls):
        """teardown any state that was previously setup with a call to setup_class
        such as deleting test data"""
        #cls.client.shutdown()
        pass

    def setup_method(self):
        """setup any state specific to all methods of the given class"""
        pass

    def teardown_method(self):
        """teardown any state that was previously setup for all methods of the given class"""
        pass

    def test_client_spawn(self):
        """
        Run astrohack_local_client with N cores and with a memory_limit of M GB to create an instance of the
        astrohack Dask client.
        """

        log_params = {
            "log_level": "DEBUG",
            "log_to_file": True,
            "log_file": "toolviper_log_file",
        }

        path = pathlib.Path(".").cwd() / "dask_test_dir"

        client = local_client(
            cores=2,
            memory_limit="8GB",
            dask_local_dir=str(path),
            log_params=log_params,
        )

        try:
            if distributed.Client.current() is None:
                raise OSError

        except OSError:
            assert False

        client.shutdown()

    def test_client_get(self):
        """
        Test the get_client() function.
        """
        from toolviper.dask.client import get_client

        client = local_client(
            cores=2,
            memory_limit="4GB",
        )

        assert get_client() == distributed.Client.current()

        client.shutdown()

    def test_cluster_get(self):
        """
        Test the get_client() function.
        """
        from toolviper.dask.client import get_cluster

        client = local_client(
            cores=2,
            memory_limit="4GB",
        )

        assert get_cluster() == distributed.Client.current().cluster

        client.shutdown()

    def test_client_thread_info(self):
        """
        Test that thread_info() function returns the values that were set in the client instantiation.
        """

        client = local_client(
            cores=2,
            memory_limit="4GB",
        )

        memory_per_thread = -1
        n_threads = 0

        # Not sure if this test is deterministic. The tests are done using github actions and, I am sure the container
        # environment will change over time. Hopefully, the test pulls out the most consistently calculated case.

        worker_items = client.cluster.scheduler_info['workers'].items()

        for worker_name, worker in worker_items:
            temp_memory_per_thread = (worker['memory_limit'] / worker['nthreads']) / (1024 ** 3)
            n_threads = n_threads + worker['nthreads']

            if (memory_per_thread == -1) or (memory_per_thread > temp_memory_per_thread):
                memory_per_thread = temp_memory_per_thread

        assert client.thread_info() == {
            "n_threads": 2,
            "memory_per_thread": memory_per_thread
        }

        client.shutdown()


    def test_client_dask_dir(self):
        """
        Run astrohack_local_client with N cores and with a memory_limit of M GB to create an instance of the
        astrohack Dask client. Check that temporary files are written to dask_local_dir.
        """

        try:

            path = pathlib.Path(".").cwd() / "dask_test_dir"

            if path.exists() is False:
                raise FileNotFoundError

        except FileNotFoundError:
            assert False

    def test_client_logger(self):
        """
        Run astrohack_local_client with N cores and with a memory_limit of M GB without any errors and the messages
        will be logged in the terminal.
        """

        files = os.listdir(".")

        try:
            for file in files:
                if re.match("^toolviper_log_file+[0-9].*log", file) is not None:
                    return

            raise FileNotFoundError

        except FileNotFoundError:
            assert False



#   Copyright 2019 AUI, Inc. Washington DC, USA
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import os
import sys
import dask
import logging

from datetime import datetime
from toolviper.utils.console import Colorize
from toolviper.utils.console import add_verbose_info

from dask.distributed import get_worker

from contextvars import ContextVar

from typing import Union

VERBOSE = True
DEFAULT = False

# global verbosity flag
verbosity: Union[ContextVar[bool], ContextVar[None]] = ContextVar(
    "message_verbosity",
    default=None
)


def set_verbosity(state: Union[None, bool] = None):
    print(f"Setting verbosity to {state}")

    verbosity.set(state)


def info(message: str, verbose: bool = False):
    logger_name = os.getenv("LOGGER_NAME")

    if verbosity.get() is True or False:
        verbose = verbosity.get()

    if verbose:
        message = add_verbose_info(message=message, color="blue")

    logger = get_logger(logger_name=logger_name)
    logger.info(message)


def log(message: str, verbose: bool = False):
    logger_name = os.getenv("LOGGER_NAME")

    if verbosity.get() is True or False:
        verbose = verbosity.get()

    if verbose:
        message = add_verbose_info(message=message, color="blue")

    logger = get_logger(logger_name=logger_name)

    logger.log(logger.level, message)


def exception(message: str, verbose: bool = False):
    logger_name = os.getenv("LOGGER_NAME")

    if verbosity.get() is True or False:
        verbose = verbosity.get()

    if verbose:
        message = add_verbose_info(message=message, color="blue")

    logger = get_logger(logger_name=logger_name)

    logger.exception(message)


def debug(message: str, verbose: bool = False):
    logger_name = os.getenv("LOGGER_NAME")

    if verbosity.get() is True or False:
        verbose = verbosity.get()

    if verbose:
        message = add_verbose_info(message=message, color="green")

    logger = get_logger(logger_name=logger_name)
    logger.debug(message)


def warning(message: str, verbose: bool = False):
    logger_name = os.getenv("LOGGER_NAME")

    if verbosity.get() is True or False:
        verbose = verbosity.get()

    if verbose:
        message = add_verbose_info(message=message, color="orange")

    logger = get_logger(logger_name=logger_name)
    logger.warning(message)


def error(message: str, verbose: bool = True):
    logger_name = os.getenv("LOGGER_NAME")

    if verbosity.get() is True or False:
        verbose = verbosity.get()

    if verbose:
        message = add_verbose_info(message=message, color="red")

    logger = get_logger(logger_name=logger_name)
    logger.error(message)


def critical(message: str, verbose: bool = True):
    logger_name = os.getenv("LOGGER_NAME")

    if verbosity.get() is True or False:
        verbose = verbosity.get()

    if verbose:
        message = add_verbose_info(message=message, color="alert")

    logger = get_logger(logger_name=logger_name)
    logger.critical(message)


class ColorLoggingFormatter(logging.Formatter):
    colorize = Colorize()

    function = " [{function}] ".format(function=colorize.blue("%(funcName)s"))
    verbose = " [{exechain}] ".format(
        exechain=colorize.blue("%(filename)s:%(lineno)s : %(module)s.%(funcName)s")
    )

    start_msg = "[{time}] ".format(time=colorize.purple("%(asctime)s"))
    middle_msg = "{level}".format(level="%(levelname)8s")
    execution_msg = " {name} [ {filename} ]: {exec_info}: ".format(
        name="%(name)10s",
        filename="%(filename)-20s",
        exec_info=colorize.blue("%(callchain)-45s"),
    )

    FORMATS = {
        logging.DEBUG: start_msg
        + colorize.green(middle_msg)
        + colorize.grey("  %(name)10s: ")
        + " %(message)s",
        logging.INFO: start_msg
        + colorize.blue(middle_msg)
        + colorize.grey("  %(name)10s: ")
        + " %(message)s ",
        logging.WARNING: start_msg
        + colorize.orange(middle_msg)
        + colorize.grey("  %(name)10s: ")
        + " %(message)s ",
        logging.ERROR: start_msg
        + colorize.red(middle_msg)
        + colorize.grey("  %(name)10s: ")
        + " %(message)s",
        logging.CRITICAL: start_msg
        + colorize.format(text=middle_msg, color=[220, 60, 20], highlight=True)
        + colorize.grey("  %(name)10s: ")
        + " %(message)s",
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


class LoggingFormatter(logging.Formatter):
    function = " [{function}] ".format(function="%(funcName)s")
    verbose = " [{exechain}] ".format(
        exechain="%(filename)s:%(lineno)s : %(module)s.%(funcName)s"
    )

    start_msg = "[{time}] ".format(time="%(asctime)s")
    middle_msg = "{level}".format(level="%(levelname)8s")
    execution_msg = " {name} [ {filename} ]: {exec_info}: ".format(
        name="%(name)10s", filename="%(filename)-20s", exec_info="%(callchain)-45s"
    )

    FORMATS = {
        logging.DEBUG: start_msg + middle_msg + "  %(name)10s: " + " %(message)s",
        logging.INFO: start_msg + middle_msg + "  %(name)10s: " + " %(message)s ",
        logging.WARNING: start_msg + middle_msg + "  %(name)10s: " + " %(message)s ",
        logging.ERROR: start_msg + middle_msg + "  %(name)10s: " + " %(message)s",
        logging.CRITICAL: start_msg + middle_msg + "  %(name)10s: " + " %(message)s",
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


def get_logger(logger_name: Union[str, None] = None):
    if logger_name is None:
        if os.getenv("LOGGER_NAME"):
            # Return default logger from env if none is specified.
            logger_name = os.getenv("LOGGER_NAME")
        else:
            logger_name = "viperlog"

    try:
        worker = get_worker()

    except ValueError:
        # Scheduler processes
        logger_dict = logging.Logger.manager.loggerDict
        if logger_name in logger_dict:
            logger = logging.getLogger(logger_name)
        else:
            # If main logger is not started using client function it defaults to printing to term.
            logger = logging.getLogger(logger_name)
            stream_handler = logging.StreamHandler(sys.stdout)
            stream_handler.setFormatter(ColorLoggingFormatter())
            logger.addHandler(stream_handler)
            logger.setLevel(logging.getLevelName("INFO"))

        return logger

    try:
        logger = worker.plugins["worker_logger"].get_logger()

        return logger

    except Exception as e:
        print("Could not load worker logger: {}".format(e))
        print(worker.plugins.keys())

        return logging.getLogger()


def setup_logger(
    logger_name: Union[str, None] = None,
    log_to_term: bool = False,
    log_to_file: bool = True,
    log_file: str = "logger",
    log_level: str = "INFO",
):
    """To set up as many loggers as you want"""
    if logger_name is None:
        logger_name = "viperlog"

    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.getLevelName(log_level))

    logger.handlers.clear()

    if log_to_term:
        stream_handler = logging.StreamHandler(sys.stdout)
        stream_handler.setFormatter(ColorLoggingFormatter())
        logger.addHandler(stream_handler)

    if log_to_file:
        log_file = log_file + datetime.today().strftime("%Y%m%d_%H%M%S") + ".log"
        log_handler = logging.FileHandler(log_file)
        log_handler.setFormatter(LoggingFormatter())
        logger.addHandler(log_handler)

    return logger


def get_worker_logger_name(logger_name: Union[str, None] = None):
    if logger_name is None:
        logger_name = "viperlog"

    return "_".join((logger_name, str(get_worker().id)))


def setup_worker_logger(
    logger_name: str,
    log_to_term: bool,
    log_to_file: bool,
    log_file: str,
    log_level: str,
    worker: dask.distributed.worker.Worker,
):
    parallel_logger_name = "_".join((logger_name, str(worker.name)))

    logger = logging.getLogger(parallel_logger_name)
    logger.setLevel(logging.getLevelName(log_level))

    if log_to_term:
        stream_handler = logging.StreamHandler(sys.stdout)
        stream_handler.setFormatter(ColorLoggingFormatter())
        logger.addHandler(stream_handler)

    if log_to_file:
        logger.info(f"log_to_file: {log_file}")
        dask.distributed.print(f"log_to_file: {log_to_file}")

        log_file = (
            log_file
            + "_"
            + str(worker.name)
            + "_"
            + datetime.today().strftime("%Y%m%d_%H%M%S")
            + "_"
            + str(worker.ip)
            + ".log"
        )
        log_handler = logging.FileHandler(log_file)
        log_handler.setFormatter(LoggingFormatter())
        logger.addHandler(log_handler)

    return logger



import numpy
import cerberus


class Protego(cerberus.Validator):
    numpy_ndarray_type = cerberus.TypeDefinition("ndarray", (numpy.ndarray,), ())
    str_type = cerberus.TypeDefinition("str", (str,), ())
    int_type = cerberus.TypeDefinition("int", (int,), ())
    tuple_type = cerberus.TypeDefinition("tuple", (tuple,), ())

    types_mapping = cerberus.Validator.types_mapping.copy()
    types_mapping["ndarray"] = numpy_ndarray_type
    types_mapping["str"] = str_type
    types_mapping["int"] = int_type
    types_mapping["tuple"] = tuple_type

    @staticmethod
    def custom_allowed_function(*args, **kwargs):
        return None

    def _validate_check_allowed_with(self, constraint, field, value):
        """Test the custom constraint

        The rule's arguments are validated against this schema:
        {'type': 'string'}
        """
        custom_constraint = self.custom_allowed_function(constraint)
        if custom_constraint is None:
            self._error(field, "Custom constraint checking function not valid")
            return

        if isinstance(value, list):
            for entry in value:
                if entry not in custom_constraint:
                    self._error(
                        field,
                        "Allowed values not satisfied: {value} != {c}".format(
                            value=value, c=custom_constraint
                        ),
                    )
        else:
            if value not in custom_constraint:
                self._error(
                    field,
                    "Allowed values not satisfied: {value} != {c}".format(
                        value=value, c=custom_constraint
                    ),
                )

    def _validate_sequence(self, constraint, field, value):
        """Test the custom constraint

        The rule's arguments are validated against this schema:
        {'type': 'list'}
        """
        if not value == constraint:
            self._error(
                field,
                "Correct sequence not satisfied: {value}!={constraint}".format(
                    value=value, constraint=constraint
                ),
            )

    def _validate_struct_type(self, constraint, field, value):
        """Test the structure type.

        constraint: True | False
        field: amount
        value: number

        The rule's arguments are validated against this schema:
        {'type': 'list'}
        """
        if isinstance(value, list):
            for entry in value:
                if not type(entry).__name__ in constraint:
                    self._error(
                        field,
                        "Must be of type:{constraint}".format(constraint=constraint),
                    )

    def register_data_type(self, data_type):
        new_data_type = cerberus.TypeDefinition(data_type.__name__, (data_type,), ())
        self.types_mapping[data_type.__name__] = new_data_type
